"""
This module contains several handy functions primarily meant for internal use.
"""

from datetime import date, datetime, timedelta
from calendar import monthrange, weekday
from time import strftime

__all__ = ('MIN_VALUES', 'MAX_VALUES', 'asint', 'asbool', 'get_actual_maximum',
           'get_date_field', 'convert_to_datetime', 'timedelta_seconds',
           'time_difference', 'datetime_ceil')


MIN_VALUES = {'year': 1970, 'month': 1, 'day': 1, 'week': 1,
              'day_of_week': 0, 'hour': 0, 'minute': 0, 'second': 0}
MAX_VALUES = {'year': 2 ** 63, 'month': 12, 'day:': 31, 'week': 53,
              'day_of_week': 6, 'hour': 23, 'minute': 59, 'second': 59}


def asint(text):
    """
    Safely converts a string to an integer, returning None if the string
    is None.

    :type text: str
    :rtype: int
    """
    if text is not None:
        return int(text)


def asbool(obj):
    """
    Interprets an object as a boolean value.

    :rtype: bool
    """
    if isinstance(obj, str):
        obj = obj.strip().lower()
        if obj in ('true', 'yes', 'on', 'y', 't', '1'):
            return True
        if obj in ('false', 'no', 'off', 'n', 'f', '0'):
            return False
        raise ValueError('Unable to interpret value "%s" as boolean' % obj)
    return bool(obj)


def get_actual_maximum(dateval, fieldname):
    """
    Retrieves the maximum applicable value for the given datetime field.

    :type dateval: datetime
    :type fieldname: str
    :rtype: int
    """
    if fieldname == 'day':
        return monthrange(dateval.year, dateval.month)[1]
    return MAX_VALUES[fieldname]


def get_date_field(dateval, fieldname):
    """
    Extracts the value of the specified field from a datetime object.

    :type dateval: datetime
    :type fieldname: str
    :rtype: int
    """
    if fieldname == 'week':
        week_str = strftime('%W', dateval.timetuple())
        return int(week_str) + 1
    if fieldname == 'day_of_week':
        return weekday(dateval.year, dateval.month, dateval.day)
    return getattr(dateval, fieldname)


def convert_to_datetime(dateval):
    """
    Converts a date object to a datetime object.
    If an actual datetime object is passed, it is returned unmodified.

    :type dateval: date
    :rtype: datetime
    """
    if isinstance(dateval, datetime):
        return dateval
    elif isinstance(dateval, date):
        return datetime.fromordinal(dateval.toordinal())
    raise TypeError('Expected date, got %s instead' % type(dateval))


def timedelta_seconds(delta):
    """
    Converts the given timedelta to seconds.

    :type delta: timedelta
    :rtype: float
    """
    return delta.days * 24 * 60 * 60 + delta.seconds + \
        delta.microseconds / 1000000.0


def time_difference(date1, date2):
    """
    Returns the time difference in seconds between the given two
    datetime objects. The difference is calculated as: date1 - date2.

    :param date1: the later datetime
    :type date1: datetime
    :param date2: the earlier datetime
    :type date2: datetime
    :rtype: float
    """
    if date1 >= date2:
        return timedelta_seconds(date1 - date2)
    return -timedelta_seconds(date2 - date1)


def datetime_ceil(dateval):
    """
    Rounds the given datetime object upwards.

    :type dateval: datetime
    """
    if dateval.microsecond > 0:
        return dateval + timedelta(seconds=1,
                                   microseconds=-dateval.microsecond)
    return dateval



"""
Triggers determine the times when a job should be executed.
"""
from datetime import datetime, timedelta
from math import ceil

from apscheduler.expressions import *
from apscheduler.util import *

__all__ = ('CronTrigger', 'DateTrigger', 'IntervalTrigger')


class CronTrigger(object):
    def __init__(self, year='*', month='*', day='*', week='*',
                 day_of_week='*', hour='*', minute='*', second='*'):
        self.fields = []
        self._compile_expressions(year, 'year')
        self._compile_expressions(month, 'month')
        self._compile_expressions(day, 'day')
        self._compile_expressions(week, 'week')
        self._compile_expressions(day_of_week, 'day_of_week')
        self._compile_expressions(hour, 'hour')
        self._compile_expressions(minute, 'minute')
        self._compile_expressions(second, 'second')

    def _compile_expressions(self, exprs, fieldname):
        def compile_single(expr):
            compilers = [AllExpression, RangeExpression]
            if fieldname == 'day_of_week':
                compilers.append(WeekdayRangeExpression)
                compilers.append(WeekdayPositionExpression)
            for compiler in compilers:
                match = compiler.value_re.match(expr)
                if match:
                    return compiler(**match.groupdict())
            raise ValueError('Unrecognized expression "%s" for field "%s"' %
                             (expr, fieldname))

        exprs = str(exprs).strip()
        if ',' in exprs:
            expr_list = exprs.split(',')
        else:
            expr_list = [exprs]
        compiled_expr_list = [compile_single(expr) for expr in expr_list]
        self.fields.append((fieldname, compiled_expr_list))

    def _set_field_value(self, dateval, fieldnum, new_value):
        """
        Sets the value of the designated field in the given datetime object.
        All less significant fields will be reset to their minimum values.

        :type dateval: datetime
        :type fieldnum: int
        :type new_value: int
        :rtype: datetime
        """

        # Fill in values for all date fields
        values = {}
        for i, field in enumerate(self.fields):
            fieldname = field[0]
            if not hasattr(dateval, fieldname):
                continue
            if i < fieldnum:
                values[fieldname] = getattr(dateval, fieldname)
            elif i > fieldnum:
                values[fieldname] = MIN_VALUES[fieldname]
            else:
                values[fieldname] = new_value

        return datetime(**values)

    def _increment_field_value(self, dateval, fieldnum):
        """
        Increments the designated field and resets all less significant fields
        to their minimum values.

        :type dateval: datetime
        :type fieldnum: int
        :type amount: int
        :rtype: tuple
        :return: a tuple containing the new date, and the number of the field
                 that was actually incremented
        """

        # If the given field is already at its maximum, or it has no counterpart
        # in a datetime object, then increment the next most significant field
        fieldname = self.fields[fieldnum][0]
        value = getattr(dateval, fieldname, None)
        if value is not None:
            maxval = get_actual_maximum(dateval, fieldname)
            if value < maxval:
                dateval = self._set_field_value(dateval, fieldnum, value + 1)
                return (dateval, fieldnum)
        return self._increment_field_value(dateval, fieldnum - 1)

    def get_next_fire_time(self, start_date):
        next_date = datetime_ceil(start_date)
        fieldnum = 0
        while fieldnum < len(self.fields):
            fieldname, expr_list = self.fields[fieldnum]
            startval = get_date_field(next_date, fieldname)

            # Calculate the smallest suitable value for this field
            nextval = None
            for expr in expr_list:
                val = expr.get_next_value(next_date, fieldname)
                if val is not None:
                    if nextval is not None:
                        nextval = min(val, nextval)
                    else:
                        nextval = val

            if nextval is None or (nextval > startval and not
                                   hasattr(datetime, fieldname)):
                # Either no valid value was found for this field,
                # or this field is a computed field and the current
                # value was not acceptable
                if fieldnum == 0:
                    # No valid values found for the first field, so give up
                    return None
                # Return to the previous field and look for the
                # next valid value
                next_date, fieldnum = self._increment_field_value(next_date,
                                                                  fieldnum)
            elif nextval > startval:
                # A valid value was found, but it was higher than the starting
                # value, so reset all less significant fields so as not to miss
                # any potential combinations
                next_date = self._set_field_value(next_date, fieldnum, nextval)
                fieldnum += 1
            else:
                # The field's value was accepted without modifications
                fieldnum += 1

        next_date = next_date - timedelta(microseconds=next_date.microsecond)
        return next_date


class DateTrigger(object):
    def __init__(self, run_date):
        self.run_date = convert_to_datetime(run_date)

    def get_next_fire_time(self, start_date):
        if self.run_date >= start_date:
            return self.run_date


class IntervalTrigger(object):
    def __init__(self, interval, repeat, start_date=None):
        if not isinstance(interval, timedelta):
            raise TypeError('interval must be a timedelta')
        if repeat < 0:
            raise ValueError('Illegal value for repeat; expected >= 0, '
                             'received %s' % repeat)

        self.interval = interval
        self.interval_length = timedelta_seconds(self.interval)
        if self.interval_length == 0:
            self.interval = timedelta(seconds=1)
            self.interval_length = 1
        self.repeat = repeat
        if start_date is None:
            self.first_fire_date = datetime.now() + self.interval
        else:
            self.first_fire_date = convert_to_datetime(start_date)
        self.first_fire_date -= timedelta(microseconds=\
                                          self.first_fire_date.microsecond)
        if repeat > 0:
            self.last_fire_date = self.first_fire_date + interval * (repeat-1)
        else:
            self.last_fire_date = None

    def get_next_fire_time(self, start_date):
        if start_date < self.first_fire_date:
            return self.first_fire_date
        if self.last_fire_date and start_date > self.last_fire_date:
            return None
        timediff_seconds = timedelta_seconds(start_date - self.first_fire_date)
        next_interval_num = int(ceil(timediff_seconds / self.interval_length))
        return self.first_fire_date + self.interval * next_interval_num



"""
This module contains the expressions applicable for CronTrigger's fields.
"""
import re
from calendar import weekday, monthrange

from apscheduler.util import *

__all__ = ['AllExpression', 'RangeExpression', 'WeekdayRangeExpression',
           'WeekdayPositionExpression']

WEEKDAYS = ['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun']


class AllExpression(object):
    value_re = re.compile(r'\*(?:/(?P<step>\d+))?$')

    def __init__(self, step=None):
        self.step = asint(step)
        if self.step == 0:
            raise ValueError('Increment must be higher than 0')

    def _get_minval(self, date, field):
        return MIN_VALUES[field]

    def _get_maxval(self, date, field):
        return get_actual_maximum(date, field)

    def get_next_value(self, date, fieldname):
        # day_of_week is not an attribute of datetime
        start = get_date_field(date, fieldname)
        minval = self._get_minval(date, fieldname)
        maxval = self._get_maxval(date, fieldname)
        start = max(start, minval)
        if not self.step:
            next = start
        else:
            distance_to_next = (self.step - (start - minval)) % self.step
            next = start + distance_to_next

        if next <= maxval:
            return next

    def __str__(self):
        if self.step:
            return '*/%d' % self.step
        return '*'


class RangeExpression(AllExpression):
    value_re = re.compile(
        r'(?P<first>\d+)(?:-(?P<last>\d+))?(?:/(?P<step>\d+))?$')

    def __init__(self, first, last=None, step=None):
        AllExpression.__init__(self, step)
        first = asint(first)
        last = asint(last)
        if last is None and step is None:
            last = first
        if last is not None and first > last:
            raise ValueError('The minimum value in a range must not be '
                             'higher than the maximum')
        self.first = first
        self.last = last

    def _get_minval(self, date, field):
        minval = AllExpression._get_minval(self, date, field)
        return max(minval, self.first)

    def _get_maxval(self, date, field):
        maxval = AllExpression._get_maxval(self, date, field)
        if self.last is not None:
            return min(maxval, self.last)
        return maxval

    def __str__(self):
        if self.last != self.first:
            if self.step:
                return '%d-%d/%d' % (self.first, self.last, self.step)
            else:
                return '%d-%d' % (self.first, self.last)
        return str(self.first)


class WeekdayRangeExpression(RangeExpression):
    value_re = re.compile(r'(?P<first>[a-z]+)(?:-(?P<last>[a-z]+))',
                          re.IGNORECASE)

    def __init__(self, first, last=None):
        try:
            first_num = WEEKDAYS.index(first.lower())
        except ValueError:
            raise ValueError('Invalid weekday name "%s"' % first)

        if last:
            try:
                last_num = WEEKDAYS.index(last.lower())
            except ValueError:
                raise ValueError('Invalid weekday name "%s"' % last)
        else:
            last_num = None

        RangeExpression.__init__(self, first_num, last_num)

    def __str__(self):
        if self.last != self.first:
            return '%s-%s' % (WEEKDAYS[self.first], WEEKDAYS[self.last])
        return WEEKDAYS[self.first]


class WeekdayPositionExpression(object):
    options = ['1st', '2nd', '3rd', '4th', '5th', 'last']
    value_re = re.compile(r'(?P<option>%s) +(?P<weekday>(?:\d+|\w+))'
                          % '|'.join(options), re.IGNORECASE)

    def __init__(self, option_name, weekday_name):
        try:
            self.option_num = self.options.index(option_name.lower())
        except ValueError:
            raise ValueError('Invalid weekday position "%s"' % option_name)

        try:
            self.weekday = WEEKDAYS.index(weekday_name.lower())
        except ValueError:
            raise ValueError('Invalid weekday name "%s"' % weekday_name)

    def get_next_value(self, date, field):
        hits = 0
        last_hit = None
        wday, last_day = monthrange(date.year, date.month)
        for day in range(date.day, last_day+1):
            wday = weekday(date.year, date.month, day)
            if wday == self.weekday:
                if hits == self.option_num:
                    return day
                hits += 1
                last_hit = day
        if self.option_num == 5:
            return last_hit

    def __str__(self):
        return '%s %s' % (self.options[self.option_num],
                          WEEKDAYS[self.weekday])



"""
This module is the main part of the library, and is the only module that
regular users should be concerned with.
"""

from threading import Thread, Event, Lock
from datetime import datetime, timedelta
from logging import getLogger

from apscheduler.util import time_difference, asbool
from apscheduler.triggers import DateTrigger, IntervalTrigger, CronTrigger


logger = getLogger(__name__)


class Job(object):
    """
    Represents a task scheduled in the scheduler.
    """

    def __init__(self, trigger, func, args, kwargs):
        self.thread = None
        self.trigger = trigger
        self.func = func
        self.args = args
        self.kwargs = kwargs
        if hasattr(func, '__name__'):
            self.name = func.__name__
        else:
            self.name = str(func)

    def run(self):
        """
        Starts the execution of this job in a separate thread.
        """
        if (self.thread and self.thread.isAlive()):
            logger.info('Skipping run of job %s (previously triggered '
                        'instance is still running)', self)
        else:
            self.thread = Thread(target=self.run_in_thread)
            self.thread.setDaemon(False)
            self.thread.start()

    def run_in_thread(self):
        """
        Runs the associated callable.
        This method is executed in a dedicated thread.
        """
        try:
            self.func(*self.args, **self.kwargs)
        except:
            logger.exception('Error executing job "%s"', self)
            raise

    def __str__(self):
        return self.name


class SchedulerShutdownError(Exception):
    """
    Thrown when attempting to use the scheduler after
    it's been shut down.
    """

    def __init__(self):
        Exception.__init__(self, 'Scheduler has already been shut down')


class SchedulerAlreadyRunningError(Exception):
    """
    Thrown when attempting to start the scheduler, but it's already running.
    """

    def __init__(self):
        Exception.__init__(self, 'Scheduler is already running')


class Scheduler(object):
    """
    This class is responsible for scheduling jobs and triggering
    their execution.
    """

    stopped = False
    thread = None
    misfire_grace_time = 1
    daemonic = True

    def __init__(self, **config):
        self.jobs = []
        self.jobs_lock = Lock()
        self.wakeup = Event()
        self.configure(config)

    def configure(self, config):
        """
        Updates the configuration with the given options.
        """
        for key, val in config.items():
            if key.startswith('apscheduler.'):
                key = key[12:]
            if key == 'misfire_grace_time':
                self.misfire_grace_time = int(val)
            elif key == 'daemonic':
                self.daemonic = asbool(val)

    def start(self):
        """
        Starts the scheduler in a new thread.
        """
        if self.thread and self.thread.isAlive():
            raise SchedulerAlreadyRunningError
        self.thread = Thread(target=self.run, name='APScheduler')
        self.thread.setDaemon(self.daemonic)
        self.thread.start()
        logger.info('Scheduler started')

    def shutdown(self, timeout=0):
        """
        Shuts down the scheduler and terminates the thread.
        Does not terminate any currently running jobs.

        :param timeout: time (in seconds) to wait for the scheduler thread to
            terminate, 0 to wait forever, None to skip waiting
        """
        if self.stopped or not self.thread.isAlive():
            return

        logger.info('Scheduler shutting down')
        self.stopped = True
        self.wakeup.set()
        if timeout is not None:
            self.thread.join(timeout)
        self.jobs = []

    def cron_schedule(self, year='*', month='*', day='*', week='*',
                      day_of_week='*', hour='*', minute='*', second='*',
                      args=None, kwargs=None):
        """
        Decorator that causes its host function to be scheduled
        according to the given parameters.
        This decorator does not wrap its host function.
        The scheduled function will be called without any arguments.
        See :meth:`add_cron_job` for more information.
        """
        def inner(func):
            self.add_cron_job(func, year, month, day, week, day_of_week, hour,
                              minute, second, args, kwargs)
            return func
        return inner

    def interval_schedule(self, weeks=0, days=0, hours=0, minutes=0, seconds=0,
                          start_date=None, repeat=0, args=None, kwargs=None):
        """
        Decorator that causes its host function to be scheduled
        for execution on specified intervals.
        This decorator does not wrap its host function.
        The scheduled function will be called without any arguments.
        Note that the default repeat value is 0, which means to repeat forever.
        See :meth:`add_delayed_job` for more information.
        """
        def inner(func):
            self.add_interval_job(func, weeks, days, hours, minutes, seconds,
                                  start_date, repeat, args, kwargs)
            return func
        return inner

    def add_job(self, trigger, func, args, kwargs):
        """
        Adds a Job to the job list and notifies the scheduler thread.

        :param trigger: trigger for the given callable
        :param args: list of positional arguments to call func with
        :param kwargs: dict of keyword arguments to call func with
        :return: the scheduled job
        :rtype: Job
        """
        if self.stopped:
            raise SchedulerShutdownError
        if not hasattr(func, '__call__'):
            raise TypeError('func must be callable')

        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        job = Job(trigger, func, args, kwargs)
        self.jobs_lock.acquire()
        try:
            self.jobs.append(job)
        finally:
            self.jobs_lock.release()
        logger.info('Added job "%s"', job)

        # Notify the scheduler about the new job
        self.wakeup.set()

        return job

    def add_date_job(self, func, date, args=None, kwargs=None):
        """
        Adds a job to be completed on a specific date and time.

        :param func: callable to run
        :param args: positional arguments to call func with
        :param kwargs: keyword arguments to call func with
        """
        trigger = DateTrigger(date)
        return self.add_job(trigger, func, args, kwargs)

    def add_interval_job(self, func, weeks=0, days=0, hours=0, minutes=0,
                         seconds=0, start_date=None, repeat=0, args=None,
                         kwargs=None):
        """
        Adds a job to be completed on specified intervals.

        :param func: callable to run
        :param weeks: number of weeks to wait
        :param days: number of days to wait
        :param hours: number of hours to wait
        :param minutes: number of minutes to wait
        :param seconds: number of seconds to wait
        :param start_date: when to first execute the job and start the
            counter (default is after the given interval)
        :param repeat: number of times the job will be run (0 = repeat
            indefinitely)
        :param args: list of positional arguments to call func with
        :param kwargs: dict of keyword arguments to call func with
        """
        interval = timedelta(weeks=weeks, days=days, hours=hours,
                             minutes=minutes, seconds=seconds)
        trigger = IntervalTrigger(interval, repeat, start_date)
        return self.add_job(trigger, func, args, kwargs)

    def add_cron_job(self, func, year='*', month='*', day='*', week='*',
                     day_of_week='*', hour='*', minute='*', second='*',
                     args=None, kwargs=None):
        """
        Adds a job to be completed on times that match the given expressions.

        :param func: callable to run
        :param year: year to run on
        :param month: month to run on (0 = January)
        :param day: day of month to run on
        :param week: week of the year to run on
        :param day_of_week: weekday to run on (0 = Monday)
        :param hour: hour to run on
        :param second: second to run on
        :param args: list of positional arguments to call func with
        :param kwargs: dict of keyword arguments to call func with
        :return: the scheduled job
        :rtype: Job
        """
        trigger = CronTrigger(year, month, day, day_of_week, hour, minute,
                              second)
        return self.add_job(trigger, func, args, kwargs)

    def is_job_active(self, job):
        """
        Determines if the given job is still on the job list.

        :return: True if the job is still active, False if not
        """
        self.jobs_lock.acquire()
        try:
            return job in self.jobs
        finally:
            self.jobs_lock.release()

    def unschedule_job(self, job):
        """
        Removes a job, preventing it from being fired any more.
        """
        self.jobs_lock.acquire()
        try:
            self.jobs.remove(job)
        finally:
            self.jobs_lock.release()
        logger.info('Removed job "%s"', job)
        self.wakeup.set()

    def unschedule_func(self, func):
        """
        Removes all jobs that would execute the given function.
        """
        self.jobs_lock.acquire()
        try:
            remove_list = [job for job in self.jobs if job.func == func]
            for job in remove_list:
                self.jobs.remove(job)
                logger.info('Removed job "%s"', job)
        finally:
            self.jobs_lock.release()

        # Have the scheduler calculate a new wakeup time
        self.wakeup.set()

    def _get_next_wakeup_time(self, now):
        """
        Determines the time of the next job execution, and removes finished
        jobs.

        :param now: the result of datetime.now(), generated elsewhere for
            consistency.
        """
        next_wakeup = None
        finished_jobs = []

        self.jobs_lock.acquire()
        try:
            for job in self.jobs:
                next_run = job.trigger.get_next_fire_time(now)
                if next_run is None:
                    finished_jobs.append(job)
                elif next_run and (next_wakeup is None or \
                                   next_run < next_wakeup):
                    next_wakeup = next_run

            # Clear out any finished jobs
            for job in finished_jobs:
                self.jobs.remove(job)
                logger.info('Removed finished job "%s"', job)
        finally:
            self.jobs_lock.release()

        return next_wakeup

    def _get_current_jobs(self):
        """
        Determines which jobs should be executed right now.
        """
        current_jobs = []
        now = datetime.now()
        start = now - timedelta(seconds=self.misfire_grace_time)

        self.jobs_lock.acquire()
        try:
            for job in self.jobs:
                next_run = job.trigger.get_next_fire_time(start)
                if next_run:
                    time_diff = time_difference(now, next_run)
                    if next_run < now and time_diff <= self.misfire_grace_time:
                        current_jobs.append(job)
        finally:
            self.jobs_lock.release()

        return current_jobs

    def run(self):
        """
        Runs the main loop of the scheduler.
        """
        self.wakeup.clear()
        while not self.stopped:
            # Execute any jobs scheduled to be run right now
            for job in self._get_current_jobs():
                logger.debug('Executing job "%s"', job)
                job.run()

            # Figure out when the next job should be run, and
            # adjust the wait time accordingly
            now = datetime.now()
            next_wakeup_time = self._get_next_wakeup_time(now)

            # Sleep until the next job is scheduled to be run,
            # or a new job is added, or the scheduler is stopped
            if next_wakeup_time is not None:
                wait_seconds = time_difference(next_wakeup_time, now)
                logger.debug('Next wakeup is due at %s (in %f seconds)',
                             next_wakeup_time, wait_seconds)
                self.wakeup.wait(wait_seconds)
            else:
                logger.debug('No jobs; waiting until a job is added')
                self.wakeup.wait()
            self.wakeup.clear()



# coding: utf-8
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


setup(
    name='APScheduler',
    version='1.2',
    description='In-process task scheduler with Cron-like capabilities',
    long_description=open('README.txt').read(),
    author='Alex Gronholm',
    author_email='apscheduler@nextday.fi',
    url='http://apscheduler.nextday.fi/',
    classifiers=[
      'Development Status :: 5 - Production/Stable',
      'Intended Audience :: Developers',
      'License :: OSI Approved :: MIT License',
      'Programming Language :: Python',
      'Programming Language :: Python :: 2.4',
      'Programming Language :: Python :: 2.5',
      'Programming Language :: Python :: 2.6',
      'Programming Language :: Python :: 3',
    ],
    keywords='scheduling cron',
    license='MIT',
    packages=['apscheduler'],
    zip_safe=True,
    test_suite='nose.collector',
    tests_require=['nose']
)
